import netgen.gui
from ngsolve import *
from netgen.geom2d import SplineGeometry
import matplotlib.pyplot as plt

#   point numbers 0, 1, ... 11
#   sub-domain numbers (1), (2), (3)
#
#    (-1,1)           (1,1)
#      3----------------2
#      |                |
#      |                |
#      |      (1)       |
#      |                |
#      |                |
#      0----------------1
#    (-1,-1)          (1,-1)

def MakeGeometry():
    geometry = SplineGeometry()

    # point coordinates ...
    pnts = [ (-1,-1), (1,-1), (1,1), (-1,1) ]
    pnums = [geometry.AppendPoint(*p) for p in pnts]

    # start-point, end-point, boundary-condition, left-domain, right-domain:
    lines = [ (0,1,"neum",1,0), (1,2,"diri",1,0), \
              (2,3,"diri",1,0), (3,0,"neum",1,0) ]

    for p1,p2,bc,left,right in lines:
        geometry.Append(["line", pnums[p1], pnums[p2]], bc=bc, leftdomain=left, rightdomain=right)

    return geometry

geo = MakeGeometry()

# generate an unstructured mesh
mesh = Mesh(geo.GenerateMesh(maxh=0.2))

# definitions of necessary coefficient functions
pt1 = Vector(2)
pt1[0] = -1/2
pt1[1] = 1/2
pt2 = Vector(2)
pt2[0] = -1/2
pt2[1] = -1/2
pt3 = Vector(2)
pt3[0] = 1/2
pt3[1] = -1/2
sigma = 1/8
c = -1/(sigma**2)
dist1 = (x-pt1[0])**2+(y-pt1[1])**2
dist2 = (x-pt2[0])**2+(y-pt2[1])**2
dist3 = (x-pt3[0])**2+(y-pt3[1])**2
uExact = exp(c*dist1) + exp(c*dist2) + exp(c*dist3)
uExactLap = 4*c*exp(c*dist1)*(c*dist1 + 1)\
    + 4*c*exp(c*dist2)*(c*dist2 + 1)\
    + 4*c*exp(c*dist3)*(c*dist3 + 1)
fRHS = -uExactLap + uExact
g1 = uExact
g2Bottom = -2*c*exp(c*dist1)*(y-pt1[1])\
    -2*c*exp(c*dist2)*(y-pt2[1])\
    -2*c*exp(c*dist3)*(y-pt3[1])
g2Left = -2*c*exp(c*dist1)*(x-pt1[0])\
    -2*c*exp(c*dist2)*(x-pt2[0])\
    -2*c*exp(c*dist3)*(x-pt3[0])

order = 1

# declare the finite element space
fes = H1(mesh, order=order, dirichlet="diri", autoupdate=True)
u = fes.TrialFunction()
v = fes.TestFunction()
print ("ndof =", fes.ndof)

# define and assemble linear and bilinear forms:
a = BilinearForm(fes, symmetric=True)
a += (grad(u)*grad(v)+u*v)*dx

f = LinearForm(fes)
f += fRHS*v*dx + g2Left*v*ds("neumLeft") + g2Bottom*v*ds("neumBottom")

gfu = GridFunction(fes)
Draw (gfu)

def SolveBVP():
    fes.Update()
    gfu.Update()
    a.Assemble()
    f.Assemble()
    gfu.Set(uExact,BND)
    c = Preconditioner(a,"local")
    c.Update()
    solvers.BVP(bf=a, lf=f, gf=gfu, pre=c)
    Redraw (blocking=True)

l = []

def CalcError():
    h = specialcf.mesh_size
    n = specialcf.normal(mesh.dim)
    def Laplace(u): return Trace(u.Operator("hesse"))
    resElement = (fRHS + Laplace(gfu) - gfu)*(fRHS + Laplace(gfu) - gfu)
    etaElement = Integrate(h*h*resElement, mesh, VOL, element_wise=True)
    etaEdge = Integrate(h*(n*(grad(gfu)-grad(gfu).Other()))**2*dx(element_boundary=True), mesh, element_wise=True)
    etaK = etaElement + (1/sqrt(2))*etaEdge
    maxerr = max(etaK)
    l.append((fes.ndof, sqrt(sum(etaK))))
    print ("ndof =", fes.ndof, "maxerr = ", maxerr)

    for el in mesh.Elements():
        mesh.SetRefinementFlag(el, etaK[el.nr] > 0.5*maxerr)

with TaskManager():
    while fes.ndof < 5000:
        SolveBVP()
        CalcError()
        mesh.Refine()

SolveBVP()
CalcError()
